@@ -2,7 +2,7 @@ use strict;
 use warnings;
 package YAML::PP::Emitter;
 
-our $VERSION = '0.023'; # VERSION
+our $VERSION = '0.025'; # VERSION
 use Data::Dumper;
 
 use YAML::PP::Common qw/
@@ -13,12 +13,14 @@ use YAML::PP::Common qw/
 /;
 
 use constant DEBUG => $ENV{YAML_PP_EMIT_DEBUG} ? 1 : 0;
+use constant DEFAULT_WIDTH => 80;
 
 sub new {
     my ($class, %args) = @_;
     my $self = bless {
         indent => $args{indent} || 2,
         writer => $args{writer},
+        width => $args{width} || DEFAULT_WIDTH,
     }, $class;
     $self->init;
     return $self;
@@ -35,6 +37,9 @@ sub clone {
 sub event_stack { return $_[0]->{event_stack} }
 sub set_event_stack { $_[0]->{event_stack} = $_[1] }
 sub indent { return $_[0]->{indent} }
+sub width { return $_[0]->{width} }
+sub line { return $_[0]->{line} }
+sub column { return $_[0]->{column} }
 sub set_indent { $_[0]->{indent} = $_[1] }
 sub writer { $_[0]->{writer} }
 sub set_writer { $_[0]->{writer} = $_[1] }
@@ -50,6 +55,8 @@ sub init {
         'tag:yaml.org,2002:' => '!!',
     });
     $self->{open_ended} = 0;
+    $self->{line} = 0;
+    $self->{column} = 0;
     $self->writer->init;
 }
 
@@ -60,6 +67,7 @@ sub mapping_start_event {
     my $last = $stack->[-1];
     my $indent = $last->{indent};
     my $new_indent = $indent;
+    my $yaml = '';
 
     my $props = '';
     my $anchor = $info->{anchor};
@@ -72,80 +80,104 @@ sub mapping_start_event {
     }
     $props = join ' ', grep defined, ($anchor, $tag);
 
-    my $column = $last->{column};
-    my $yaml = '';
+    my $flow = $last->{flow} || 0;
+    $flow++ if ($info->{style} || 0) eq YAML_FLOW_MAPPING_STYLE;
+
     my $newline = 0;
-    if ($last->{type} eq 'DOC') {
-        if ($props) {
-            $newline = 1;
-            $yaml .= $last->{column} ? ' ' : $indent;
-            $yaml .= "$props";
-        }
-        if ($last->{newline}) {
-                $newline = 1;
-        }
-        else {
-            if ($props) {
-                $newline = 1;
+    if ($flow > 1) {
+        if ($last->{type} eq 'SEQ') {
+            if ($last->{newline}) {
+                $yaml .= ' ';
             }
-        }
-    }
-    else {
-        $new_indent .= ' ' x $self->indent;
-        if ($last->{newline}) {
-            $yaml .= "\n";
-            $last->{column} = 0;
-            $last->{newline} = 0;
-        }
-        if ($props) {
-            $newline = 1;
-        }
-        if ($last->{type} eq 'MAPVALUE') {
-            $newline = 1;
-        }
-        else {
-            $yaml .= $last->{column} ? ' ' : $indent;
-            $last->{newline} = 0;
-            if ($last->{type} eq 'SEQ') {
-                $yaml .= '-';
-            }
-            elsif ($last->{type} eq 'MAP') {
-                $yaml .= "?";
-                $last->{type} = 'COMPLEX';
-            }
-            elsif ($last->{type} eq 'COMPLEX') {
-                $yaml .= ":";
-                $last->{type} = 'COMPLEXVALUE';
-            }
-            elsif ($last->{type} eq 'COMPLEXVALUE') {
-                $yaml .= ":";
-                $last->{type} = 'MAP';
+            if ($last->{index} == 0) {
+                $yaml .= "[";
             }
             else {
-                die "Unexpected";
+                $yaml .= ",";
             }
         }
+        elsif ($last->{type} eq 'MAP') {
+            if ($last->{newline}) {
+                $yaml .= ' ';
+            }
+            if ($last->{index} == 0) {
+                $yaml .= "{";
+            }
+            else {
+                $yaml .= ",";
+            }
+        }
+        elsif ($last->{type} eq 'MAPVALUE') {
+            if ($last->{index} == 0) {
+                die "Should not happen (index 0 in MAPVALUE)";
+            }
+            $yaml .= ": ";
+        }
         if ($props) {
-            $yaml .= " $props";
+            $yaml .= " $props ";
+        }
+        $new_indent .= ' ' x $self->indent;
+    }
+    else {
+        if ($last->{type} eq 'DOC') {
+            $newline = $last->{newline};
+        }
+        else {
+            if ($last->{newline}) {
+                $yaml .= "\n";
+                $last->{column} = 0;
+            }
+            if ($last->{type} eq 'MAPVALUE') {
+                $new_indent .= ' ' x $self->indent;
+                $newline = 1;
+            }
+            else {
+                $new_indent = $indent;
+                if (not $props and $self->indent == 1) {
+                    $new_indent .= ' ' x 2;
+                }
+                else {
+                    $new_indent .= ' ' x $self->indent;
+                }
+
+                if ($last->{column}) {
+                    my $space = $self->indent > 1 ? ' ' x ($self->indent - 1) : ' ';
+                    $yaml .= $space;
+                }
+                else {
+                    $yaml .= $indent;
+                }
+                if ($last->{type} eq 'SEQ') {
+                    $yaml .= '-';
+                }
+                elsif ($last->{type} eq 'MAP') {
+                    $yaml .= "?";
+                    $last->{type} = 'COMPLEX';
+                }
+                elsif ($last->{type} eq 'COMPLEXVALUE') {
+                    $yaml .= ":";
+                }
+                else {
+                    die "Should not happen ($last->{type} in mapping_start)";
+                }
+                $last->{column} = 1;
+            }
+            $last->{newline} = 0;
+        }
+        if ($props) {
+            $yaml .= $last->{column} ? ' ' : $indent;
+            $yaml .= $props;
             $newline = 1;
         }
     }
-    if (length $yaml) {
-        $column = substr($yaml, -1) eq "\n" ? 0 : 1;
-    }
-    $self->writer->write($yaml);
+    $self->_write($yaml);
     my $new_info = {
         index => 0, indent => $new_indent, info => $info,
         newline => $newline,
-        column => $column,
+        column => $self->column,
+        flow => $flow,
     };
-    if (($info->{style} || '') eq YAML_FLOW_MAPPING_STYLE) {
-#        $new_info->{type} = 'FLOWMAP';
-        $new_info->{type} = 'MAP';
-    }
-    else {
-        $new_info->{type} = 'MAP';
-    }
+    $new_info->{type} = 'MAP';
     push @{ $stack }, $new_info;
     $last->{index}++;
     $self->{open_ended} = 0;
@@ -157,26 +189,32 @@ sub mapping_end_event {
     my $stack = $self->event_stack;
 
     my $last = pop @{ $stack };
-    my $column = $last->{column};
     if ($last->{index} == 0) {
         my $indent = $last->{indent};
         my $zero_indent = $last->{zero_indent};
         if ($last->{zero_indent}) {
             $indent .= ' ' x $self->indent;
         }
-        if ($last->{column}) {
-            $self->writer->write(" {}\n");
+        if ($self->column) {
+            $self->_write(" {}\n");
         }
         else {
-            $self->writer->write("$indent\{}\n");
+            $self->_write("$indent\{}\n");
         }
-        $column = 0;
+    }
+    elsif ($last->{flow}) {
+        my $yaml = "}";
+        if ($last->{flow} == 1) {
+            $yaml .= "\n";
+        }
+        $self->_write("$yaml");
     }
     $last = $stack->[-1];
-    $last->{column} = $column;
+    $last->{column} = $self->column;
     if ($last->{type} eq 'SEQ') {
     }
     elsif ($last->{type} eq 'MAP') {
+        $last->{type} = 'MAPVALUE';
     }
     elsif ($last->{type} eq 'MAPVALUE') {
         $last->{type} = 'MAP';
@@ -198,7 +236,6 @@ sub sequence_start_event {
     my $new_indent = $indent;
     my $yaml = '';
 
-    my $writer = $self->writer;
     my $props = '';
     my $anchor = $info->{anchor};
     my $tag = $info->{tag};
@@ -210,66 +247,108 @@ sub sequence_start_event {
     }
     $props = join ' ', grep defined, ($anchor, $tag);
 
+    my $flow = $last->{flow} || 0;
+    $flow++ if $flow or ($info->{style} || 0) eq YAML_FLOW_SEQUENCE_STYLE;
     my $newline = 0;
     my $zero_indent = 0;
-    if ($last->{type} eq 'DOC') {
-        $newline = $last->{newline};
+    if ($flow > 1) {
+        if ($last->{type} eq 'SEQ') {
+            if ($last->{newline}) {
+                $yaml .= ' ';
+            }
+            if ($last->{index} == 0) {
+                $yaml .= "[";
+            }
+            else {
+                $yaml .= ",";
+            }
+        }
+        elsif ($last->{type} eq 'MAP') {
+            if ($last->{newline}) {
+                $yaml .= ' ';
+            }
+            if ($last->{index} == 0) {
+                $yaml .= "{";
+            }
+            else {
+                $yaml .= ",";
+            }
+        }
+        elsif ($last->{type} eq 'MAPVALUE') {
+            if ($last->{index} == 0) {
+                die "Should not happen (index 0 in MAPVALUE)";
+            }
+            $yaml .= ": ";
+        }
+        if ($props) {
+            $yaml .= " $props ";
+        }
+        $new_indent .= ' ' x $self->indent;
     }
     else {
-        if ($last->{newline}) {
-            $yaml .= "\n";
-            $last->{column} = 0;
-            $last->{newline} = 0;
-        }
-        if ($last->{type} eq 'MAPVALUE') {
-            $zero_indent = 1;
-            $newline = 1;
+        if ($last->{type} eq 'DOC') {
+            $newline = $last->{newline};
         }
         else {
+            if ($last->{newline}) {
+                $yaml .= "\n";
+                $last->{column} = 0;
+            }
+            if ($last->{type} eq 'MAPVALUE') {
+                $zero_indent = 1;
+                $newline = 1;
+            }
+            else {
+                if (not $props and $self->indent == 1) {
+                    $new_indent .= ' ' x 2;
+                }
+                else {
+                    $new_indent .= ' ' x $self->indent;
+                }
+                if ($last->{column}) {
+                    my $space = $self->indent > 1 ? ' ' x ($self->indent - 1) : ' ';
+                    $yaml .= $space;
+                }
+                else {
+                    $yaml .= $indent;
+                }
+                if ($last->{type} eq 'SEQ') {
+                    $yaml .= "-";
+                }
+                elsif ($last->{type} eq 'MAP') {
+                    $last->{type} = 'COMPLEX';
+                    $zero_indent = 1;
+                    $yaml .= "?";
+                }
+                elsif ($last->{type} eq 'COMPLEXVALUE') {
+                    $yaml .= ":";
+                    $zero_indent = 1;
+                }
+                else {
+                    die "Should not happen ($last->{type} in sequence_start)";
+                }
+                $last->{column} = 1;
+            }
+            $last->{newline} = 0;
+        }
+        if ($props) {
             $yaml .= $last->{column} ? ' ' : $indent;
-            if ($last->{type} eq 'SEQ') {
-                $new_indent .= ' ' x $self->indent;
-                $yaml .= "-";
-            }
-            elsif ($last->{type} eq 'MAP') {
-                $new_indent .= ' ' x $self->indent;
-                $yaml .= "?";
-                $last->{type} = 'COMPLEX';
-                $zero_indent = 1;
-            }
-            elsif ($last->{type} eq 'COMPLEXVALUE') {
-                $new_indent .= ' ' x $self->indent;
-                $yaml .= ":";
-                $zero_indent = 1;
-            }
-            $last->{column} = 1;
+            $yaml .= $props;
+            $newline = 1;
         }
     }
-    if ($props) {
-        $newline = 1;
-        $yaml .= $last->{column} ? ' ' : $indent;
-        $yaml .= $props;
-    }
-    $self->writer->write($yaml);
+    $self->_write($yaml);
     $last->{index}++;
-    my $column = $last->{column};
-    if (length $yaml) {
-        $column = substr($yaml, -1) eq "\n" ? 0 : 1;
-    }
     my $new_info = {
         index => 0,
         indent => $new_indent,
         info => $info,
         zero_indent => $zero_indent,
         newline => $newline,
-        column => $column,
+        column => $self->column,
+        flow => $flow,
     };
-    if (($info->{style} || '') eq YAML_FLOW_SEQUENCE_STYLE) {
-        $new_info->{type} = 'FLOWSEQ';
-    }
-    else {
-        $new_info->{type} = 'SEQ';
-    }
+    $new_info->{type} = 'SEQ';
     push @{ $stack }, $new_info;
     $self->{open_ended} = 0;
 }
@@ -280,23 +359,32 @@ sub sequence_end_event {
     my $stack = $self->event_stack;
 
     my $last = pop @{ $stack };
-    my $column = $last->{column};;
     if ($last->{index} == 0) {
         my $indent = $last->{indent};
         my $zero_indent = $last->{zero_indent};
         if ($last->{zero_indent}) {
             $indent .= ' ' x $self->indent;
         }
-        my $yaml .= $last->{column} ? ' ' : $indent;
-        $yaml .= "[]\n";
-        $self->writer->write("$yaml");
-        $column = 0;
+        my $yaml .= $self->column ? ' ' : $indent;
+        $yaml .= "[]";
+        if ($last->{flow} < 2) {
+            $yaml .= "\n";
+        }
+        $self->_write($yaml);
+    }
+    elsif ($last->{flow}) {
+        my $yaml = "]";
+        if ($last->{flow} == 1) {
+            $yaml .= "\n";
+        }
+        $self->_write($yaml);
     }
     $last = $stack->[-1];
-    $last->{column} = $column;
+    $last->{column} = $self->column;
     if ($last->{type} eq 'SEQ') {
     }
     elsif ($last->{type} eq 'MAP') {
+        $last->{type} = 'MAPVALUE';
     }
     elsif ($last->{type} eq 'MAPVALUE') {
         $last->{type} = 'MAP';
@@ -404,6 +492,7 @@ sub scalar_event {
     my $last = $stack->[-1];
     my $indent = $last->{indent};
     my $value = $info->{value};
+    my $flow = $last->{flow};
 
     my $props = '';
     my $anchor = $info->{anchor};
@@ -416,21 +505,26 @@ sub scalar_event {
     }
     $props = join ' ', grep defined, ($anchor, $tag);
 
-
     my $style = $info->{style};
     DEBUG and local $Data::Dumper::Useqq = 1;
     $value = '' unless defined $value;
-    if (not $style and $value eq '') {
-        $style = YAML_SINGLE_QUOTED_SCALAR_STYLE;
+    my $first = substr($value, 0, 1);
+
+    if ($value eq '') {
+        if ($flow and $last->{type} ne 'MAPVALUE' and $last->{type} ne 'MAP') {
+            $style = YAML_SINGLE_QUOTED_SCALAR_STYLE;
+        }
+        elsif (not $style) {
+            $style = YAML_SINGLE_QUOTED_SCALAR_STYLE;
+        }
+    }
+    # no control characters anywhere
+    elsif ($value =~ m/[$control_re]/) {
+        $style = YAML_DOUBLE_QUOTED_SCALAR_STYLE;
     }
     $style ||= YAML_PLAIN_SCALAR_STYLE;
 
-    my $first = substr($value, 0, 1);
-    # no control characters anywhere
-    if ($style ne YAML_DOUBLE_QUOTED_SCALAR_STYLE and $value =~ m/[$control_re]/) {
-        $style = YAML_DOUBLE_QUOTED_SCALAR_STYLE;
-    }
-    elsif ($style eq YAML_SINGLE_QUOTED_SCALAR_STYLE) {
+    if ($style eq YAML_SINGLE_QUOTED_SCALAR_STYLE) {
         if ($value =~ m/ \n/ or $value =~ m/\n / or $value =~ m/^\n/ or $value =~ m/\n$/) {
             $style = YAML_DOUBLE_QUOTED_SCALAR_STYLE;
         }
@@ -439,24 +533,44 @@ sub scalar_event {
         }
     }
     elsif ($style eq YAML_LITERAL_SCALAR_STYLE or $style eq YAML_FOLDED_SCALAR_STYLE) {
+        if ($value eq '') {
+            $style = YAML_DOUBLE_QUOTED_SCALAR_STYLE;
+        }
+        elsif ($flow) {
+            # no block scalars in flow
+            if ($value =~ tr/\n//) {
+                $style = YAML_DOUBLE_QUOTED_SCALAR_STYLE;
+            }
+            else {
+                $style = YAML_SINGLE_QUOTED_SCALAR_STYLE;
+            }
+        }
     }
     elsif ($style eq YAML_PLAIN_SCALAR_STYLE) {
-        if ($value =~ m/[$escape_re_without_lb]/) {
+        if (not length $value) {
+        }
+        elsif ($value =~ m/[$escape_re_without_lb]/) {
             $style = YAML_DOUBLE_QUOTED_SCALAR_STYLE;
         }
         elsif ($value eq "\n") {
             $style = YAML_DOUBLE_QUOTED_SCALAR_STYLE;
         }
-        elsif ($value =~ m/\n/) {
-            $style = YAML_LITERAL_SCALAR_STYLE;
+        elsif ($value !~ tr/ //c) {
+            $style = YAML_SINGLE_QUOTED_SCALAR_STYLE;
+        }
+        elsif ($value !~ tr/ \n//c) {
+            $style = YAML_DOUBLE_QUOTED_SCALAR_STYLE;
+        }
+        elsif ($value =~ tr/\n//) {
+            $style = $flow ? YAML_DOUBLE_QUOTED_SCALAR_STYLE : YAML_LITERAL_SCALAR_STYLE;
         }
         elsif ($forbidden_first{ $first }) {
             $style = YAML_SINGLE_QUOTED_SCALAR_STYLE;
         }
-        elsif (substr($value, 0, 3) =~ m/^(?:---|\.\.\.)/) {
+        elsif ($flow and $value =~ tr/,[]{}//) {
             $style = YAML_SINGLE_QUOTED_SCALAR_STYLE;
         }
-        elsif (substr($value, 0, 2) =~ m/^(?:[:?-] )/) {
+        elsif (substr($value, 0, 3) =~ m/^(?:---|\.\.\.)/) {
             $style = YAML_SINGLE_QUOTED_SCALAR_STYLE;
         }
         elsif ($value =~ m/: /) {
@@ -469,26 +583,22 @@ sub scalar_event {
             $style = YAML_SINGLE_QUOTED_SCALAR_STYLE;
         }
         elsif ($value =~ m/[^\x20-\x3A\x3B-\x7E\x85\xA0-\x{D7FF}\x{E000}-\x{FEFE}\x{FF00}-\x{FFFD}\x{10000}-\x{10FFFF}]/) {
-            # TODO exclude ,[]{} in flow collections
             $style = YAML_SINGLE_QUOTED_SCALAR_STYLE;
         }
-        else {
-            $style = YAML_PLAIN_SCALAR_STYLE;
-        }
-    }
-
-    my $open_ended = 0;
-    if ($style eq YAML_PLAIN_SCALAR_STYLE) {
-        if ($forbidden_first_plus_space{ $first }) {
+        elsif ($forbidden_first_plus_space{ $first }) {
             if (length ($value) == 1 or substr($value, 1, 1) =~ m/^\s/) {
                 $style = YAML_SINGLE_QUOTED_SCALAR_STYLE;
             }
         }
     }
-
-    if (($style eq YAML_LITERAL_SCALAR_STYLE or $style eq YAML_FOLDED_SCALAR_STYLE) and $value eq '') {
-        $style = YAML_DOUBLE_QUOTED_SCALAR_STYLE;
+    if ($style eq YAML_SINGLE_QUOTED_SCALAR_STYLE and not $info->{style}) {
+        if ($value =~ tr/'// and $value !~ tr/"//) {
+            $style = YAML_DOUBLE_QUOTED_SCALAR_STYLE;
+        }
     }
+
+    my $open_ended = 0;
+
     if ($style eq YAML_PLAIN_SCALAR_STYLE) {
         $value =~ s/\n/\n\n/g;
     }
@@ -512,6 +622,7 @@ sub scalar_event {
         if ($value =~ m/\A\n* +/) {
             $indicators .= $self->indent;
         }
+        my $indent = $indent . ' ' x $self->indent;
         if ($value !~ m/\n\z/) {
             $indicators .= '-';
             $value .= "\n";
@@ -520,7 +631,7 @@ sub scalar_event {
             $indicators .= '+';
             $open_ended = 1;
         }
-        $value =~ s/^(?=.)/$indent  /gm;
+        $value =~ s/^(?=.)/$indent/gm;
         $value = "|$indicators\n$value";
     }
     elsif ($style eq YAML_FOLDED_SCALAR_STYLE) {
@@ -532,6 +643,7 @@ sub scalar_event {
         if ($value =~ m/\A\n* +/) {
             $indicators .= $self->indent;
         }
+        my $indent = $indent . ' ' x $self->indent;
         if ($lines[-1] eq '') {
             pop @lines;
             $eol = 1;
@@ -543,7 +655,7 @@ sub scalar_event {
         for my $i (0 .. $#lines) {
             my $line = $lines[ $i ];
             if (length $line) {
-                $value .= "$indent  $line\n";
+                $value .= "$indent$line\n";
             }
             if ($i != $#lines) {
                 $value .= "\n";
@@ -566,74 +678,125 @@ sub scalar_event {
     }
     my $multiline = ($style eq YAML_LITERAL_SCALAR_STYLE or $style eq YAML_FOLDED_SCALAR_STYLE);
     my $newline = 0;
-    my $column = $last->{column};
-    if ($last->{type} eq 'MAP' or $last->{type} eq 'SEQ') {
-        if ($last->{index} == 0 and $last->{newline}) {
-            $yaml .= "\n";
-            $last->{column} = 0;
-            $last->{newline} = 0;
-        }
-    }
-    if ($last->{type} eq 'MAP') {
-
+    if ($flow) {
+        $indent = 0;
         if ($props and not length $value) {
             $pvalue .= ' ';
         }
-        my $new_event = 'MAPVALUE';
-        $yaml .= $last->{column} ? ' ' : $indent;
-        if ($multiline) {
-            # oops, a complex key
-            $yaml .= "? ";
-            $new_event = 'COMPLEXVALUE';
-        }
-        if (not $multiline) {
-            $pvalue .= ":";
-        }
-        $last->{type} = $new_event;
-    }
-    else {
-        if ($last->{type} eq 'MAPVALUE') {
-            $last->{type} = 'MAP';
-        }
-        elsif ($last->{type} eq 'DOC') {
-        }
-        else {
-            $yaml .= $last->{column} ? ' ' : $indent;
-            if ($last->{type} eq 'COMPLEXVALUE') {
-                $last->{type} = 'MAP';
-                $yaml .= ":";
-            }
-            elsif ($last->{type} eq 'COMPLEX') {
-                $yaml .= ": ";
-            }
-            elsif ($last->{type} eq 'SEQ') {
-                $yaml .= "-";
+        if ($last->{type} eq 'SEQ') {
+            if ($last->{index} == 0) {
+                if ($self->column) {
+                    $yaml .= ' ';
+                }
+                $yaml .= "[";
             }
             else {
-                die "Unexpected";
-            }
-            $last->{column} = 1;
-        }
-
-        if (length $pvalue) {
-            if ($last->{column}) {
-                $pvalue = " $pvalue";
+                $yaml .= ", ";
             }
         }
-        if (not $multiline) {
-            $pvalue .= "\n";
+        elsif ($last->{type} eq 'MAP') {
+            if ($last->{index} == 0) {
+                if ($self->column) {
+                    $yaml .= ' ';
+                }
+                $yaml .= "{";
+            }
+            else {
+                $yaml .= ", ";
+            }
+            $last->{type} = 'MAPVALUE';
         }
+        elsif ($last->{type} eq 'MAPVALUE') {
+            if ($last->{index} == 0) {
+                die "Should not happen (index 0 in MAPVALUE)";
+            }
+            $yaml .= ": ";
+            $last->{type} = 'MAP';
+        }
+        if ($self->column + length $pvalue > $self->width) {
+            $yaml .= "\n";
+            $yaml .= $last->{indent};
+            $yaml .= ' ' x $self->indent;
+        }
+        $yaml .= $pvalue;
+    }
+    else {
+        if ($last->{type} eq 'MAP' or $last->{type} eq 'SEQ') {
+            if ($last->{index} == 0 and $last->{newline}) {
+                $yaml .= "\n";
+                $last->{column} = 0;
+                $last->{newline} = 0;
+            }
+        }
+        my $space = ' ';
+        if ($last->{type} eq 'MAP') {
+
+            if ($last->{column}) {
+                my $space = $self->indent > 1 ? ' ' x ($self->indent - 1) : ' ';
+                $yaml .= $space;
+            }
+            else {
+                $yaml .= $indent;
+            }
+            if ($props and not length $value) {
+                $pvalue .= ' ';
+            }
+            my $new_event = 'MAPVALUE';
+            $last->{type} = $new_event;
+            if ($multiline) {
+                # oops, a complex key
+                $yaml .= "? ";
+                $new_event = 'COMPLEXVALUE';
+                $last->{type} = $new_event;
+            }
+            if (not $multiline) {
+                $pvalue .= ":";
+            }
+        }
+        else {
+            if ($last->{type} eq 'MAPVALUE') {
+                $last->{type} = 'MAP';
+            }
+            elsif ($last->{type} eq 'DOC') {
+            }
+            else {
+                if ($last->{column}) {
+                    my $space = $self->indent > 1 ? ' ' x ($self->indent - 1) : ' ';
+                    $yaml .= $space;
+                }
+                else {
+                    $yaml .= $indent;
+                }
+                if ($last->{type} eq 'COMPLEXVALUE') {
+                    $last->{type} = 'MAP';
+                    $yaml .= ":";
+                }
+                elsif ($last->{type} eq 'SEQ') {
+                    $yaml .= "-";
+                }
+                else {
+                    die "Should not happen ($last->{type} in scalar_event)";
+
+                }
+                $last->{column} = 1;
+            }
+
+            if (length $pvalue) {
+                if ($last->{column}) {
+                    $pvalue = "$space$pvalue";
+                }
+            }
+            if (not $multiline) {
+                $pvalue .= "\n";
+            }
+        }
+        $yaml .= $pvalue;
     }
-    $yaml .= $pvalue;
 
-    $column = $last->{column};
     $last->{index}++;
     $last->{newline} = $newline;
-    if (length $yaml) {
-        $column = substr($yaml, -1) eq "\n" ? 0 : 1;
-    }
-    $last->{column} = $column;
-    $self->writer->write($yaml);
+    $self->_write($yaml);
+    $last->{column} = $self->column;
     $self->{open_ended} = $open_ended;
 }
 
@@ -643,6 +806,7 @@ sub alias_event {
     my $stack = $self->event_stack;
     my $last = $stack->[-1];
     my $indent = $last->{indent};
+    my $flow = $last->{flow};
 
     my $alias = '*' . $info->{value};
 
@@ -655,43 +819,82 @@ sub alias_event {
         }
     }
     $yaml .= $last->{column} ? ' ' : $indent;
-    if ($last->{type} eq 'MAP') {
-        $yaml .= "$alias :";
-        $last->{type} = 'MAPVALUE';
-    }
-    else {
-
-        if ($last->{type} eq 'MAPVALUE') {
-            $last->{type} = 'MAP';
-        }
-        elsif ($last->{type} eq 'DOC') {
-            # TODO an alias at document level isn't actually valid
-        }
-        else {
-            if ($last->{type} eq 'COMPLEXVALUE') {
-                $last->{type} = 'MAP';
-                $yaml .= ": ";
-            }
-            elsif ($last->{type} eq 'COMPLEX') {
-                $yaml .= ": ";
-            }
-            elsif ($last->{type} eq 'SEQ') {
-                $yaml .= "- ";
+    if ($flow) {
+        my $space = '';
+        if ($last->{type} eq 'SEQ') {
+            if ($last->{index} == 0) {
+                if ($flow == 1) {
+                    $yaml .= ' ';
+                }
+                $yaml .= "[";
             }
             else {
-                die "Unexpected";
+                $yaml .= ", ";
             }
         }
-        $yaml .= "$alias\n";
+        elsif ($last->{type} eq 'MAP') {
+            if ($last->{index} == 0) {
+                if ($flow == 1) {
+                    $yaml .= ' ';
+                }
+                $yaml .= "{";
+            }
+            else {
+                $yaml .= ", ";
+            }
+            $last->{type} = 'MAPVALUE';
+            $space = ' ';
+        }
+        elsif ($last->{type} eq 'MAPVALUE') {
+            if ($last->{index} == 0) {
+                die 23;
+                if ($flow == 1) {
+                    $yaml .= ' ';
+                }
+                $yaml .= "{";
+            }
+            else {
+                $yaml .= ": ";
+            }
+            $last->{type} = 'MAP';
+        }
+        $yaml .= "$alias$space";
+    }
+    else {
+        if ($last->{type} eq 'MAP') {
+            $yaml .= "$alias :";
+            $last->{type} = 'MAPVALUE';
+        }
+        else {
+
+            if ($last->{type} eq 'MAPVALUE') {
+                $last->{type} = 'MAP';
+            }
+            elsif ($last->{type} eq 'DOC') {
+                # TODO an alias at document level isn't actually valid
+            }
+            else {
+                if ($last->{type} eq 'COMPLEXVALUE') {
+                    $last->{type} = 'MAP';
+                    $yaml .= ": ";
+                }
+                elsif ($last->{type} eq 'COMPLEX') {
+                    $yaml .= ": ";
+                }
+                elsif ($last->{type} eq 'SEQ') {
+                    $yaml .= "- ";
+                }
+                else {
+                    die "Unexpected";
+                }
+            }
+            $yaml .= "$alias\n";
+        }
     }
 
-    $self->writer->write("$yaml");
+    $self->_write("$yaml");
     $last->{index}++;
-    my $column = $last->{column};
-    if (length $yaml) {
-        $column = substr($yaml, -1) eq "\n" ? 0 : 1;
-    }
-    $last->{column} = $column;
+    $last->{column} = $self->column;
     $self->{open_ended} = 0;
 }
 
@@ -699,25 +902,23 @@ sub document_start_event {
     DEBUG and warn __PACKAGE__.':'.__LINE__.": +++ document_start_event\n";
     my ($self, $info) = @_;
     my $newline = 0;
-    my $column = 0;
     my $implicit = $info->{implicit};
     if ($info->{version_directive}) {
         if ($self->{open_ended}) {
-            $self->writer->write("...\n");
+            $self->_write("...\n");
         }
-        $self->writer->write("%YAML $info->{version_directive}->{major}.$info->{version_directive}->{minor}\n");
+        $self->_write("%YAML $info->{version_directive}->{major}.$info->{version_directive}->{minor}\n");
         $self->{open_ended} = 0;
         $implicit = 0; # we need ---
     }
     unless ($implicit) {
         $newline = 1;
-        $self->writer->write("---");
-        $column = 1;
+        $self->_write("---");
     }
     $self->set_event_stack([
         {
         type => 'DOC', index => 0, indent => '', info => $info,
-        newline => $newline, column => $column,
+        newline => $newline, column => $self->column,
         }
     ]);
 }
@@ -727,7 +928,7 @@ sub document_end_event {
     my ($self, $info) = @_;
     $self->set_event_stack([]);
     if ($self->{open_ended} or not $info->{implicit}) {
-        $self->writer->write("...\n");
+        $self->_write("...\n");
         $self->{open_ended} = 0;
     }
     else {
@@ -764,6 +965,26 @@ sub finish {
     $self->writer->finish;
 }
 
+sub _write {
+    my ($self, $yaml) = @_;
+    return unless length $yaml;
+    my @lines = split m/\n/, $yaml, -1;
+    my $newlines = @lines - 1;
+    $self->{line} += $newlines;
+    if (length $lines[-1]) {
+        if ($newlines) {
+            $self->{column} = length $lines[-1];
+        }
+        else {
+            $self->{column} += length $lines[-1];
+        }
+    }
+    else {
+        $self->{column} = 0;
+    }
+    $self->writer->write($yaml);
+}
+
 1;
 
 __END__
